import numpy as np

class gginf():
    """
    Implements an G/G/Inf queue and computes several performance metrics 
    
    The G/G/Inf queue has interarrival and service times that follow any probability 
    distribution, with arr_rate and ser_rate (respectively), and the number of servers, 
    the system capacity and the population are infinite.

    The class uses the fact that the infinite number of servers simplifies may of the 
    metrics to obtain closed-form results for many average performance metrics.
    """
    # arrival rate
    arr_rate:np.float64

    # service rate
    ser_rate:np.float64

    # initializer 
    def __init__(self, arr_rate:np.float64, ser_rate:np.float64):
        """
        Creates an G/G/Inf queue with infinite servers, arr_rate arrival rate and ser_rate service rate 
        """
        self.arr_rate = arr_rate
        self.ser_rate = ser_rate

    def utilization(self)->np.float64:
        """
        Computes the mean server utilization
        
        As the queue has infinite servers, the utilization is always 0
        """
        return 0

    def mean_number_entities_queue(self)-> np.float64:
        """
        Computes the mean number of entities in queue in steady state

        As the queue has infinite servers, there is never a line in queue
        """
        return 0
        
    def mean_number_entities_service(self)-> np.float64:
        """
        Computes the mean number of entities in service in steady state
        """
        return self.arr_rate/self.ser_rate
    
    def mean_number_entities(self)-> np.float64:
        """
        Computes the mean number of entities in the system in steady state
        """
        return self.mean_number_entities_queue() + self.mean_number_entities_service()
        
    def mean_time_system(self)-> np.float64:
        """
        Computes the mean time in the system in steady state 
        """
        return self.mean_time_service()+self.mean_time_queue()
        
    def mean_time_queue(self)->np.float64:
        """
        Computes the mean time in the queue in steady state
        """
        return 0
    def mean_time_service(self)->np.float64:
        """
        Computes the mean time in service
        """
        return 1/self.ser_rate

    def is_stable(self)-> bool:
        """
        Returns True if the queue is stable, False otherwise
        
        This queue is always stable
        """
        return True

    




import os.path
import sys
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import simpy
import numpy as np
import pandas as pd
from scipy.stats import norm, entropy, skew, kurtosis
from jmarkov.phase.fit.moments_ctph2 import moments_ctph2
from jmarkov.phase.ctph import ctph
from jmarkov.queue.phph1 import phph1
import itertools
import time



# ---------------------------
# LAS OPERACIONES SE ENCUENTRAN ENTRE "Debug1", "Debug2" y "Debug3"
# ---------------------------

np.random.seed(123)
def generate_correlated_exponentials(n, mean, rho):
    cov = rho * np.ones((n, n)) + (1 - rho) * np.eye(n)
    mv_norm = np.random.multivariate_normal(mean=[0]*n, cov=cov)
    uniforms = norm.cdf(mv_norm)
    exponentials = -mean * np.log(uniforms)
    return exponentials


def estimate_rho_from_corr_matrix(corr_matrix):
    n = corr_matrix.shape[0]
    upper_tri = corr_matrix[np.triu_indices(n, k=1)]
    return np.mean(upper_tri)


def simulate_queue(N, INTER_ARRIVAL_TIME, TASK_MEAN_DURATION, RHO, SIM_TIME):
    env = simpy.Environment()
    server = simpy.Resource(env, capacity=1)
    all_task_durations, waiting_times, system_times = [], [], []

    def job(env, job_id, N, task_mean, rho, server):
        arrival_time = env.now
        durations = generate_correlated_exponentials(N, task_mean, rho)
        all_task_durations.append(durations)
        with server.request() as req:
            yield req
            start_service = env.now
            waiting_times.append(start_service - arrival_time)
            service_time = durations.max()
            yield env.timeout(service_time)
            completion_time = env.now
            system_times.append(completion_time - arrival_time)

    def job_generator(env, N, task_mean, interarrival, rho, server):
        job_id = 0
        while True:
            yield env.timeout(np.random.exponential(interarrival))
            env.process(job(env, job_id, N, task_mean, rho, server))
            job_id += 1

    env.process(job_generator(env, N, TASK_MEAN_DURATION, INTER_ARRIVAL_TIME, RHO, server))
    env.run(until=SIM_TIME)

    return np.array(all_task_durations), np.array(waiting_times), np.array(system_times)


def fit_phase_type_and_compare(task_matrix, waiting_times, system_times, N, INTER_ARRIVAL_TIME):

    empirical_corr_matrix = np.corrcoef(task_matrix.T)
    estimated_rho = estimate_rho_from_corr_matrix(empirical_corr_matrix)

    mean_all_tasks = task_matrix.mean()
    l1 = mean_all_tasks
    l2 = task_matrix.var() + l1*l1
    l12 = estimated_rho
    b = np.sqrt((1 - l12) * (l2 - l1*l1))
    m1 = l1 - b
    m2 = l2 - l1*l1 + m1*m1 - b*b
    #print("Llegué0")
    fitter = moments_ctph2(m1, m2)
    #print("Llegué0")
    print(m1)
    print(m2)
    PH = fitter.get_ph()
    #print("Llegué10")

    lda = 1 / INTER_ARRIVAL_TIME
    alpha = np.array([1])
    T = np.array([[-lda]])
    IAT = ctph(alpha, T)

    mu1 = 1/b
    num_tasks = N
    beta = np.zeros(num_tasks + 2)
    beta[0], beta[1] = PH.alpha[0], PH.alpha[1]
    S = np.zeros((num_tasks + 2, num_tasks + 2))
    S[0:2, 0:2] = PH.T
    S[0:2, 2:3] = -PH.T @ np.ones((2, 1))
    for i in range(num_tasks - 1):
        S[i+2, i+2] = -(num_tasks - i) * mu1
        S[i+2, i+3] = (num_tasks - i) * mu1
    S[num_tasks+1, num_tasks+1] = -mu1
    #print("Llegué1")
    ST = ctph(beta, S)
    #print("Llegué2")
    q = phph1(IAT, ST)



    WT = q.wait_time_dist()

    RT = q.resp_time_dist()



    # Compare empirical vs theoretical CDFs
    #t_max = max(system_times.max())
    t_max = system_times.max()
    dt = 0.05
    ts = np.arange(0, t_max, dt)
    print("Debug1")
    start_time = time.time()
    #theo_wait_cdf = np.array([WT.cdf(t) for t in ts])
    print("Wait time CDF comp: %s seconds" % (time.time() - start_time))

    start_time = time.time()
    theo_wait_cdf = WT.cdf(0, t_max, dt)
    print("Wait time CDF comp 2: %s seconds" % (time.time() - start_time))

    #diff_wait = np.max(theo_wait_cdf - theo_wait_cdf2)
    #print(f'Max diff wait: {diff_wait}')

    print("Debug2")
    start_time = time.time()
    #theo_resp_cdf = np.array([RT.cdf(t) for t in ts])
    print("Resp time CDF comp: %s seconds" % (time.time() - start_time))

    start_time = time.time()
    theo_resp_cdf = RT.cdf(0, t_max, dt)
    print("Resp time CDF comp: %s seconds" % (time.time() - start_time))

    #diff_resp = np.max(theo_resp_cdf - theo_resp_cdf2)
    #print(f'Max diff resp: {diff_resp}')


    print("Debug3")
    #print("A ver acá:")
    #print(len(theo_wait_cdf))
    #print(len(waiting_times))
    counts, bin_edges = np.histogram(waiting_times, bins=len(ts), density=False)

    # Convert to cumulative distribution
    cdf = np.cumsum(counts)
    cdf = cdf / cdf[-1]  # Normalize to 1


    # Evaluate the CDF at points ts using interpolation
    cdf_values = np.interp(ts, bin_edges[1:], cdf)  # Use upper edges for alignment


    #print(len(cdf_values))






    counts, bin_edges = np.histogram(system_times, bins=len(ts), density=False)

    # Convert to cumulative distribution
    cdf = np.cumsum(counts)
    cdf = cdf / cdf[-1]  # Normalize to 1

    # Evaluate the CDF at points ts using interpolation
    cdf_values2 = np.interp(ts, bin_edges[1:], cdf)  # Use upper edges for alignment
    #print("AQUÍ 1")
    #print(ts)
    #print("AQUÍ 2")
    #print(bin_edges[1:])
    #print("AQUÍ 3")
    #print(cdf)
    #print("AQUÍ 4")
    #print(cdf_values2)


    #print(len(cdf_values2))
    #print(len(theo_resp_cdf))


    ###################### Corregir las métricas #######
    ################### Utilizar métodos del RT ######################
    def metrics(sim_data, theo_cdf,indicador):
        hist, edges = np.histogram(sim_data, bins=100, range=(0, t_max), density=True)
        bin_centers = (edges[:-1] + edges[1:]) / 2
        theo_pdf = np.gradient(theo_cdf, dt)
        theo_pdf = np.interp(bin_centers, ts, theo_pdf)
        eps = 1e-10
        p, q = hist + eps, theo_pdf + eps
        kl_div = entropy(p, q)
        if(indicador==1):
          return {
            'KL': kl_div,
            'Mean1': np.mean(sim_data),
            'Std1': np.std(sim_data),
            'Skew1': skew(sim_data),
            'Kurtosis1': kurtosis(sim_data),
            'Mean2': np.mean(theo_cdf),
            'Std2': np.std(theo_cdf),
            'Skew2': skew(theo_cdf),
            'Kurtosis2': kurtosis(theo_cdf),
            'hist': cdf_values,
            'theo_pdf': theo_wait_cdf
        }
        else:
          return {
            'KL': kl_div,
            'Mean1': np.mean(sim_data),
            'Std1': np.std(sim_data),
            'Skew1': skew(sim_data),
            'Kurtosis1': kurtosis(sim_data),
            'Mean2': np.mean(theo_cdf),
            'Std2': np.std(theo_cdf),
            'Skew2': skew(theo_cdf),
            'Kurtosis2': kurtosis(theo_cdf),
            'hist': cdf_values2,
            'theo_pdf': theo_resp_cdf
        }



    wait_metrics = metrics(waiting_times, theo_wait_cdf,1)
    resp_metrics = metrics(system_times, theo_resp_cdf,2)

    return wait_metrics, resp_metrics


# ---------------------------
# Parameter grid
# ---------------------------
#N_values = [5,10,50,100]
N_values = [1000]

#INTER_ARRIVAL_VALUES = [3,4,5]
TASK_MEAN_VALUES = [1]
#RHO_VALUES = [0.3,0.5,0.7,0.9]
RHO_VALUES = [0.3]
SIM_TIME_VALUES = [10000]
#utilizacion=[0.1,0.3,0.5,0.7,0.9]
utilizacion=[0.5]

#N_values = [3]
#INTER_ARRIVAL_VALUES = [3]
#TASK_MEAN_VALUES = [0.2]
#RHO_VALUES = [0.5]
#SIM_TIME_VALUES = [10000]



results = []

for N, ut, TM, RHO, SIM_T in itertools.product(
        N_values, utilizacion, TASK_MEAN_VALUES, RHO_VALUES, SIM_TIME_VALUES):
    HN=sum(1/i for i in range(1, N+1))
    #print(IAT)
    #print(HN)
    IAT=1/(ut/HN)
    start_time = time.time()
    task_matrix, waits, systems = simulate_queue(N, IAT, TM, RHO, SIM_T)
    print("Sim time: %s seconds" % (time.time() - start_time))
    
    print("Simulación terminada")
    wait_m, resp_m= fit_phase_type_and_compare(task_matrix, waits, systems, N, IAT)
    print("Ajuste terminado")
    results.append({
        'N': N,
        'InterArrival': IAT,
        'TaskMean': TM,
        'Rho': RHO,
        'SimTime': SIM_T,
        'KL_wait': wait_m['KL'],
        'Mean_wait1': wait_m['Mean1'],
        'Mean_wait2': wait_m['Mean2'],
        'Std_wait1': wait_m['Std1'],
        'Std_wait2': wait_m['Std2'],
        'Skew_wait1': wait_m['Skew1'],
        'Skew_wait2': wait_m['Skew2'],
        'Kurt_wait1': wait_m['Kurtosis1'],
        'Kurt_wait2': wait_m['Kurtosis2'],
        'KL_resp': resp_m['KL'],
        'Mean_resp1': resp_m['Mean1'],
        'Mean_resp2': resp_m['Mean2'],
        'Std_resp1': resp_m['Std1'],
        'Std_resp2': resp_m['Std2'],
        'Skew_resp1': resp_m['Skew1'],
        'Skew_resp2': resp_m['Skew2'],
        'Kurt_resp1': resp_m['Kurtosis1'],
        'Kurt_resp2': resp_m['Kurtosis2'],
        "counts_w": wait_m['hist'],
        "theo_w": wait_m['theo_pdf'],
        "counts_r": resp_m['hist'],
        "theo_r": resp_m['theo_pdf']
    })

df = pd.DataFrame(results)
print(df)



#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.phph1 import phph1
from jmarkov.phase.ctph import ctph

# arrival rate
lda = 0.1
alpha = np.array([1])
T = np.array([[-lda]])
IAT = ctph(alpha, T)

# job service time distribution
mean_task_service_time = 1
num_tasks = 10
task_service_rate = 1/mean_task_service_time
beta = np.zeros(num_tasks)
beta[0] = 1
S = np.zeros((num_tasks,num_tasks))
for i in range(num_tasks-1):
    S[i,i] = -(num_tasks-i)*task_service_rate
    S[i,i+1] = (num_tasks-i)*task_service_rate
S[num_tasks-1,num_tasks-1] = -task_service_rate
print(f'beta: {beta}')
print(f'S: {S}')
ST = ctph(beta, S)
q = phph1(IAT,ST) 

print("Distributions of the number of entities in system:")
print(q.number_entities_dist())

L = q.mean_number_entities()
print("Mean number of entities in system:")
print(L)

Lq = q.mean_number_entities_queue()
print("Mean number of entities in queue:")
print(Lq)

W = q.mean_time_system()
print("Mean time in system:")
print(W)

Wq = q.mean_time_queue()
print("Mean time in queue:")
print(Wq)

WT = q.wait_time_dist()
print("Waiting time distribution:")
print(f'alpha: {WT.alpha}')
print(f'T: {WT.T}')




import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.ctmc import ctmc


class TestSteadyState(unittest.TestCase):
    def test_ctmc_steady_state(self):
        Q = np.array([[-4, 1, 3], [2, -5, 3], [1, 2, -3]])
        mc = ctmc(Q)
        assert_allclose(mc.steady_state(), [0.25, 0.25, 0.5],err_msg="should be [0.25, 0.25, 0.5]")

class TestIsErgodic(unittest.TestCase):
    def test_ctmc_isergodic(self):
        Q = np.array([[-4, 1, 3], [2, -5, 3], [1, 2, -3]])
        mc = ctmc(Q)
        self.assertTrue(mc.is_ergodic())

class TestFirstPassageTime(unittest.TestCase):
    def test_ctmc_first_passage_time(self):
        Q = np.array([[-3, 2, 1], [2, -5, 3], [1, 1, -2]])
        mc = ctmc(Q)
        assert_allclose(mc.first_passage_time(0), [[0.71428571], [0.85714286]], err_msg="should be [0.71428571, 0.85714286]")

class TestAbsorbtionTimes(unittest.TestCase):
    def test_absorbtion_times(self):
        Q = np.array([[-4, 1, 3], [2, -5, 3], [0, 0, 0]])
        states = np.array([0,1,2])
        mc = ctmc(Q, states)
        self.assertAlmostEqual(mc.absorbtion_times(target=0,start=1)[0][0],0.1111111)

class TestAbsorbtionProbabilities(unittest.TestCase):
    def test_absortion_probabilities(self):
        A = ctmc(np.array([[-5, 1, 3,1], [3, -10, 3,4], [0, 0, 0,0],[0,0,0,0]]),
         states = np.array([0,1,2,3]))
        self.assertAlmostEqual(A.absorbtion_probabilities(target=2,start=0)[0][0], 0.702127659574468)

if __name__ == '__main__':
    unittest.main()


import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.phase.ctph import ctph


class TestSteadyState(unittest.TestCase):
    def test_ctph_pdf(self):
        alpha = np.array([0.5, 0.5])
        T = np.array([[-2, 0], [0, -5]])
        ph = ctph(alpha, T)
        assert_allclose(ph.pdf(0.2), 1.590018648964245,err_msg="should be 1.590018648964245")

class TestIsErgodic(unittest.TestCase):
    def test_ctph_cdf(self):
        alpha = np.array([0.5, 0.5])
        T = np.array([[-2, 0], [0, -5]])
        ph = ctph(alpha, T)
        assert_allclose(ph.cdf(0.2), 0.480900256396459,err_msg="should be 0.480900256396459")

if __name__ == '__main__':
    unittest.main()


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.sdp.dtsdp import dtsdp


#Vector de épocas
E = np.array([i for i in range(1,4)])
# Variables
S = np.array(["Excelente","Bueno","Promedio","Malo"]) # Estado de la máquina al inicio de la semana
# Decisiones
A = np.array(["Reemplazar","No Reemplazar"])
# Retornos Inmediatos
R = np.zeros((len(E),len(S),len(A)))
# Recorremos sobre las épocas
for t in range(len(E)):
    # Recorremos sobre los estados:
    for s_index, i in enumerate(S):
        # Recorremos sobre las decisiones:
        for a_index, a in enumerate(A):
            if i=='Excelente' and a=='Reemplazar':
                R[t,s_index,a_index] = -1000
            elif i=='Excelente' and a=='No Reemplazar':
                R[t,s_index,a_index] = 100
            elif(i=="Bueno" and a=="Reemplazar"):
                R[t,s_index,a_index]=-100
            elif(i=="Bueno" and a=="No Reemplazar"):
                R[t,s_index,a_index]=80
            elif(i=="Promedio" and a=="Reemplazar"):
                R[t,s_index,a_index]=-100
            elif(i=="Promedio" and a=="No Reemplazar"):
                R[t,s_index,a_index]=50
            elif(i=="Malo" and a=="Reemplazar"):
                R[t,s_index,a_index]=-100
            elif(i=="Malo" and a=="No Reemplazar"):
                R[t,s_index,a_index]=10

# Matrices de transición
matNoReemplazar = np.array([[0.7,0.3,0,0],
                          [0,0.7,0.3,0],
                          [0,0,0.6,0.4],
                          [0,0,0,1]])

matReemplazar = np.array([[0,0,0,0],
                          [0.7,0.3,0,0],
                          [0.7,0.3,0,0],
                          [0.7,0.3,0,0]])
probs = {}
for t in E:  # Iterar sobre cada época
    decisiones_dict = {}
    for posA, a in enumerate(A):
        if a == "Reemplazar":
            decisiones_dict[a] = matReemplazar
        elif a == "No Reemplazar":
            decisiones_dict[a] = matNoReemplazar
    probs[t] = decisiones_dict

sdpMaquinas = dtsdp(E,S,A,probs,R,0.9)
print(sdpMaquinas.solve(minimize = False))


import numpy as np
from jmarkov.ctbd import ctbd
from jmarkov.phase.ctph import ctph

class phph1():
    """
    Implements a PH/PH/1 queue and computes steady state metrics 
    
    The PH/PH/1 queue has phase type interarrival times, with parameters (alpha,T) of size ma,
    phase type service times, with parameters (beta,S) of size ms, 
    and 1 server.

    The class builds a quasi-birth-death chain that models this queue and uses 
    the steady state probability distribution of the chain to compute
    measures of performance such as mean number of entities in the system,
    in queue, in service, and the mean time in the system, in queue, and in service.
    """

    # phase-type representation of the inter-arrival time distribution
    IAT:ctph

    # phase-type representation of the service time distribution
    ST:ctph

    # steady state probabilities
    probs:np.array

    # number of entries in the steady state vector to compute
    n:int

    # phase-type representation of the waiting time distribution
    WT:ctph

    # phase-type representation of the response time distribution
    RT:ctph


    # initializer 
    def __init__(self, IAT:ctph, ST:ctph):
        """
        Creates a PH/PH/1 queue with 1 server, IAT inter-arrival times and ST service times
        """
        self.IAT=IAT
        self.ST = ST

        mean_iat = IAT.expected_value()
        mean_st = ST.expected_value()
        rho = mean_st/mean_iat

        if rho >= 1:
            raise ValueError(f'The load {rho} exceeds one, que queue is unstable')   

        self.n=1000
        self.probs = np.full(self.n+1, np.nan)
        

    def _solve_mc(self, verbose=False):
        """
        Solves the underlying Markov chain to obtain the queue-length distribution

        Employs a method by Latouche and Ramaswami* especially devised for the PH/PH/1 queue,
        where the blocks in the underlying quasi-birth-death Markov chain are further divided in 
        subblocks.

        G. Latouche and V. Ramaswami. The PH/PH/1 queue at epochs of queue size change. 
        Queueing Systems, 25:97–114, 1997.
        """

        # arrivals
        ma = self.IAT.n_phases
        alpha = self.IAT.alpha
        # convert alpha to row vector
        alpha = alpha[np.newaxis,:]
        T = self.IAT.T
        t = -np.sum(T,axis=1)
        # convert t to column vector
        t = t[:,np.newaxis]
        avgt = self.IAT.expected_value()

        # service
        ms = self.ST.n_phases
        beta = self.ST.alpha
        # convert beta to row vector
        beta = beta[np.newaxis,:]
        S = self.ST.T
        s = -np.sum(S,axis=1)
        # convert s to column vector
        s = s[:,np.newaxis]
        avgs = self.ST.expected_value()

        mtot = ms*ma
        rho = avgs/avgt
        if rho >= 1:
            raise ValueError(f'The load {rho} exceeds one, que queue is unstable')   


        # Compute classic QBD blocks A0, A1 and A2
        A0 = np.kron(np.outer(t,alpha), np.eye(ms))
        A1 = np.kron(T,np.eye(ms))+ np.kron(np.eye(ma),S)

        # Compute QBD blocks in approach Latouche & Ramaswami
        invmA1 = np.linalg.inv(-A1)

        A0pp = np.kron(alpha,np.eye(ms))@invmA1@np.kron(t,np.eye(ms))
        A0mp = np.kron(np.eye(ma),beta)@invmA1@np.kron(t,np.eye(ms))
        A2pm = np.kron(alpha,np.eye(ms))@invmA1@np.kron(np.eye(ma),s)
        A2mm = np.kron(np.eye(ma),beta)@invmA1@np.kron(np.eye(ma),s)

        #A0n = [A0pp, np.zeros((ms,ma))], [A0mp, np.zeros((ma,ma))]]
        A0n = np.concatenate((
                np.concatenate((A0pp, np.zeros((ms,ma))), axis=1),
                np.concatenate((A0mp, np.zeros((ma,ma))), axis=1)
                ),
                axis=0
            )
        #A2n = [[np.zeros((ms,ms)), A2pm], [np.zeros((ma,ms)), A2mm]]
        A2n = np.concatenate((
                np.concatenate((np.zeros((ms,ms)), A2pm), axis=1),
                np.concatenate((np.zeros((ma,ms)), A2mm), axis=1)
                ),
                axis=0
            )

        # Compute matrix Gamma: NE corner of matrix G
        itB0 = A0n
        itB2 = A2n
        Gamma = itB2[0:ms,ms:ms+ma]
        itT = itB0
        check=1
        numit = 1
        while check > 10e-14 and numit < 10:
            itA1 = itB0@itB2 + itB2@itB0
            itB0 = np.linalg.inv(np.eye(ma+ms)-itA1)@itB0@itB0
            itB2 = np.linalg.inv(np.eye(ma+ms)-itA1)@itB2@itB2
            tmp = itT@itB2
            Gamma = Gamma + tmp[0:ms,ms:ms+ma]
            itT = itT@itB0
            check = np.linalg.norm(np.ones((ms,1))-Gamma@np.ones((ma,1)))
            numit=numit+1
            if verbose==True:
                print(f'Check after {numit} iterations: {check}')

        Gm = np.linalg.inv(np.eye(ma)-A0mp@Gamma)@A2mm
        R_Gam = A0pp@np.linalg.inv(np.eye(ms)-Gamma@A0mp)


        # Compute pi and queue length distribution
        Gstar = invmA1@np.kron(np.eye(ma),s@beta) + invmA1@np.kron(t,np.eye(ms))@Gamma@Gm@np.kron(np.eye(ma),beta)
        Rstar = A0@np.linalg.inv(-A1-A0@Gstar)

        # Compute pi_0
        pi = np.kron((1-rho)*np.linalg.inv(beta@Gamma@np.linalg.inv(-T)@np.ones((ma,1)))@beta@Gamma@np.linalg.inv(-T),beta)
        # Compute pi_1,...
        sumpi=np.sum(pi)
        numit=0
        while sumpi < 1-10**(-10) and numit < 1+self.n:
            tmp = pi[numit,:]@Rstar; #compute pi_(numit+1)
            pi = np.append(pi,[tmp],axis=0)
            pi[numit+1] = tmp
            numit=numit+1
            sumpi=sumpi+sum(pi[numit,:])
            if verbose==True:
                print(f'Accumulated mass after {numit} iterations: {sumpi}')

        self.probs[0:pi.shape[0]] = np.sum(pi,axis=1)
        self.probs[pi.shape[0]:self.n+1] = 0
        pi=np.reshape(pi, pi.shape[0]*pi.shape[1])
        if numit == 1+self.n:
            raise Warning(f'Maximum Number of Components {numit-1} reached')
        
        # wait time distribution
        sigtilde = np.linalg.inv(beta@np.linalg.inv(S)@np.ones((ms,1)))@beta@np.linalg.inv(S)
        # turn 2d array into 1d array 
        sigtilde = sigtilde[0]
        Delta = np.diag(sigtilde)
        
        wait_T = np.linalg.inv(Delta)@np.transpose(S+R_Gam@s@beta)@Delta
        theta = (-beta@np.linalg.inv(S)@np.ones((ms,1)))@np.transpose(s)@Delta
        D = np.linalg.inv(Delta)@np.transpose(R_Gam)@Delta
        wait_alpha = (1-np.linalg.inv(beta@np.linalg.inv(np.eye(ms)-R_Gam)@np.ones((ms,1)))@beta@np.ones((ms,1)))@np.linalg.inv(theta@D@np.ones((ms,1)))@theta@D
        # turn 2d array into 1d array 
        wait_alpha = wait_alpha[0]
        self.WT = ctph(wait_alpha, wait_T)
        mw = len(wait_alpha)
        # build ph representation of the response time distribution
        norm_beta = (1-wait_alpha.sum())*beta
        norm_beta = norm_beta[0]
        resp_alpha = np.concatenate((wait_alpha, norm_beta), axis=0)
        resp_exit = -wait_T@np.ones((mw,1))
        resp_T = np.concatenate((
                np.concatenate((wait_T, resp_exit@beta ), axis=1),
                np.concatenate((np.zeros((ms,mw)), S), axis=1)
                ),
                axis=0
            )
        self.RT = ctph(resp_alpha,resp_T)
    
    def number_entities_dist(self)-> float:
        """
        Computes the distribution of the number of entities in the system in steady state
        
        A quasi-birth-death chain is built and its stationary probability distribution is used 
        to compute the distribution of the number of entities in the system in steady state
        """
        if self.is_stable():
            if np.isnan(self.probs).any():
                self._solve_mc()
            return self.probs
        else:
            print('Unstable queue')
            return 0.0
    

    def mean_number_entities(self)-> np.float64:
        """
        Computes the mean number of entities in the system in steady state
        
        A quasi-birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the system in steady state
        """
        if self.is_stable():
            if np.isnan(self.probs).any():
                self._solve_mc()
            mean_num = 0 
            for i in range(self.n):
                mean_num += self.probs[i]*i

            return mean_num
        else:
            print('Unstable queue')
            return 0


    def mean_number_entities_queue(self)-> np.float64:
        """
        Computes the mean number of entities in queue in steady state
        
        A quasi-birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in queue in steady state
        """
        if self.is_stable():
            if np.isnan(self.probs).any():
                self._solve_mc()
            mean_num = 0 
            for i in range(1, self.n):
                mean_num += self.probs[i]*(i-1)

            return mean_num
        else:
            print('Unstable queue')
            return 0

    def mean_number_entities_service(self)-> np.float64:
        """
        Computes the mean number of entities in service in steady state
        
        A birth-death chain is built and its stattionary probability distribution is used 
        to compute the mean number of entities in the system in steady state
        """
        if self.is_stable():
            if np.isnan(self.probs).any():
                self._solve_mc()
            return 1 - self.probs[0]
        else:
            print('Unstable queue')
            return 0
        
    def wait_time_dist(self)-> ctph:
        """
        Computes the phase-type representation of the distribution of the waiting time in steady state
        
        A quasi-birth-death chain is built and its stationary probability distribution is used 
        to compute the phase-type representation of the distribution of the waiting time in steady state
        """
        if self.is_stable():
            if np.isnan(self.probs).any():
                self._solve_mc()
            return self.WT
        else:
            print('Unstable queue')
            return 0.0
        
    def resp_time_dist(self)-> ctph:
        """
        Computes the phase-type representation of the distribution of the response time in steady state
        
        A quasi-birth-death chain is built and its stationary probability distribution is used 
        to compute the phase-type representation of the distribution of the waiting time in steady state,
        which is then combined with the service time distribution to obtain the response time distribution
        """
        if self.is_stable():
            if np.isnan(self.probs).any():
                self._solve_mc()
            return self.RT
        else:
            print('Unstable queue')
            return 0.0


    def mean_time_system(self)-> np.float64:
        """
        Computes the mean time in the system in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the system in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        if self.is_stable():
            L = self.mean_number_entities()
            arr_rate = 1/self.IAT.expected_value()
            return L/arr_rate
        else:
            print('Unstable queue')
            return 0
    
    def mean_time_queue(self)-> np.float64:
        """
        Computes the mean time in the queue in steady state
        
        A birth-death chain is built and its stationary probability distribution is used 
        to compute the mean number of entities in the queue in steady state, which is 
        then used with Little's Law to obtain the mean time in the system in steady state
        """
        if self.is_stable():
            Lq = self.mean_number_entities_queue()
            arr_rate = 1/self.IAT.expected_value()
            return Lq/arr_rate
        else:
            print('Unstable queue')
            return 0
    
    def mean_time_service(self)-> np.float64:
        """
        Computes the mean time in service
        
        A simple relation is used to obtain the mean service time
        """
        if self.is_stable():
            return self.ST.expected_value()
        else:
            print('Unstable queue')
            return 0
    

    def utilization(self)-> np.float64:
        """
        Computes the mean server utilization
        
        The mean server utilization is computed as the ratio between the arrival rate
        and the service rate times the number of servers
        """
        rho = self.ST.expected_value()/self.IAT.expected_value()
        return rho
    
    def is_stable(self)-> bool:
        """
        Returns True if the queue is stable, False otherwise
        
        This queue is stable if the arrival rate is smaller than the 
        maximum service rate
        """
        return self.ST.expected_value() < self.IAT.expected_value() 
        


#!/usr/bin/env python
import sys
import os.path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import simpy
import numpy as np
from scipy.stats import norm

# ----------------
# Parameters
# ----------------
N = 3
INTER_ARRIVAL_TIME = 1.5
#INTER_ARRIVAL_TIME = 5
TASK_MEAN_DURATION = 0.8
print(f'base load: {TASK_MEAN_DURATION/INTER_ARRIVAL_TIME}')
SIM_TIME = 10000*5
RHO = 0.5

# To store results
all_task_durations = []
waiting_times = []
system_times = []

def generate_correlated_exponentials(n, mean, rho):
    cov = rho * np.ones((n, n)) + (1 - rho) * np.eye(n)
    mv_norm = np.random.multivariate_normal(mean=[0]*n, cov=cov)
    uniforms = norm.cdf(mv_norm)
    exponentials = -mean * np.log(uniforms)
    return exponentials

def job(env, job_id, N, task_mean, rho, server):
    arrival_time = env.now
    durations = generate_correlated_exponentials(N, task_mean, rho)
    all_task_durations.append(durations)

    with server.request() as req:
        yield req
        start_service = env.now
        waiting_times.append(start_service - arrival_time)

        # service time = max of tasks (parallel execution)
        service_time = durations.max()
        yield env.timeout(service_time)

        completion_time = env.now
        system_times.append(completion_time - arrival_time)

def job_generator(env, N, task_mean, interarrival, rho, server):
    job_id = 0
    while True:
        yield env.timeout(np.random.exponential(interarrival))
        env.process(job(env, job_id, N, task_mean, rho, server))
        job_id += 1

# ----------------
# Run Simulation
# ----------------
env = simpy.Environment()
server = simpy.Resource(env, capacity=1)  # single server like in PH/PH/1
env.process(job_generator(env, N, TASK_MEAN_DURATION, INTER_ARRIVAL_TIME, RHO, server))
env.run(until=SIM_TIME)


# Convert to matrix
task_matrix = np.vstack(all_task_durations)



def estimate_rho_from_corr_matrix(corr_matrix):
    n = corr_matrix.shape[0]
    upper_tri = corr_matrix[np.triu_indices(n, k=1)]
    return np.mean(upper_tri)

empirical_corr_matrix = np.corrcoef(task_matrix.T)
estimated_rho = estimate_rho_from_corr_matrix(empirical_corr_matrix)

mean_all_tasks = task_matrix.mean()
l1=mean_all_tasks
l2=task_matrix.var()+l1*l1
l12=estimated_rho
print(l1)
print(l2)
print(l12)

b=np.sqrt((1-l12)*(l2-l1*l1))
print(b)

m1=l1-b
print(m1)

m2=l2-l1*l1+m1*m1-b*b

print(m2)

from jmarkov.phase.fit.moments_ctph2 import moments_ctph2
from jmarkov.phase.ctph import ctph
print(m1)
print(m2)
fitter = moments_ctph2(m1,m2)
PH = fitter.get_ph()



from jmarkov.queue.phph1 import phph1
from jmarkov.phase.ctph import ctph

lda = 1 / INTER_ARRIVAL_TIME
alpha = np.array([1])
T = np.array([[-lda]])
IAT = ctph(alpha, T)

print(f'mean IAT: {IAT.expected_value()}')


num_tasks = N
mu1=1/b
beta = np.zeros(num_tasks + 2)
beta[0] = PH.alpha[0]
beta[1] = PH.alpha[1]
S = np.zeros((num_tasks + 2, num_tasks + 2))
S[0:2, 0:2] = PH.T
S[0:2, 2:3] = -PH.T@np.ones((2,1))

for i in range(num_tasks-1):
  S[i+2, i+2] = -(num_tasks - i) * mu1
  S[i+2, i + 3] = (num_tasks - i) * mu1


S[num_tasks+1, num_tasks+1] = -mu1


print(f'\nbeta: {beta}')
print(f'S:\n{S}')
ST = ctph(beta, S)

print(f'mean ST: {ST.expected_value()}')
print(f'load: {ST.expected_value()/IAT.expected_value()}')
q = phph1(IAT, ST)




WT = q.wait_time_dist()
print("Waiting time distribution:")
print(f'alpha: {WT.alpha}')
print(f'T: {WT.T}')

###########################################
RT= q.resp_time_dist()
print("Response time distribution:")
print(f'alpha: {RT.alpha}')
print(f'T: {RT.T}')






import matplotlib.pyplot as plt
import numpy as np
# ----------------
# Histogram of simulated waiting times
# ----------------
bins = 50




counts, edges = np.histogram(waiting_times, bins=bins, density=False)

# cumulative counts
cum_counts = np.cumsum(counts)

# normalize to turn it into a CDF (range [0,1])
cdf = cum_counts / cum_counts[-1]

counts=cdf



#counts/=(counts.sum()*dt)
bin_centers = (edges[:-1] + edges[1:]) / 2

# ----------------
# Theoretical distribution from jMarkov
# ----------------
t_max = max(max(waiting_times), 20)
dt = 0.05
ts = np.arange(0, t_max, dt)

#theo_pdf = np.array([WT.pdf(t) for t in ts])
theo_pdf = np.array([WT.cdf(t) for t in ts])
#theo_pdf /= np.sum(theo_pdf) * dt  # normalize
#theo_pdf /= INTENSIDAD_TRAFICO
# ----------------
# Plot
# ----------------
plt.figure(figsize=(8,5))
plt.bar(bin_centers, counts, width=edges[1]-edges[0], alpha=0.6,
        label="Simulated Waiting Times", color="orange", edgecolor="black")
plt.plot(ts, theo_pdf, 'b-', linewidth=2, label="jMarkov PDF")

plt.xlabel("Waiting Time")
plt.ylabel("Density")
plt.title("Comparison of Waiting Time Distributions")
plt.legend()
plt.grid(True, linestyle="--", alpha=0.6)
plt.tight_layout()
plt.show()













import matplotlib.pyplot as plt
import numpy as np
#max_task_durations = np.max(task_matrix, axis=1)  # From your simulation
#max_task_durations_ = np.pad(max_task_durations, (0, 8), constant_values=0)
sim_response_times=system_times

# ----------------
# Histogram of simulated waiting times
# ----------------
bins = 50




counts, edges = np.histogram(sim_response_times, bins=bins, density=False)

# cumulative counts
cum_counts = np.cumsum(counts)

# normalize to turn it into a CDF (range [0,1])
cdf = cum_counts / cum_counts[-1]

counts=cdf



#counts/=(counts.sum()*dt)
bin_centers = (edges[:-1] + edges[1:]) / 2

# ----------------
# Theoretical distribution from jMarkov
# ----------------
t_max = max(max(sim_response_times), 20)
dt = 0.05
ts = np.arange(0, t_max, dt)

#theo_pdf = np.array([WT.pdf(t) for t in ts])
theo_pdf = np.array([RT.cdf(t) for t in ts])
#theo_pdf /= np.sum(theo_pdf) * dt  # normalize
#theo_pdf /= INTENSIDAD_TRAFICO
# ----------------
# Plot
# ----------------
plt.figure(figsize=(8,5))
plt.bar(bin_centers, counts, width=edges[1]-edges[0], alpha=0.6,
        label="Simulated Service Times", color="orange", edgecolor="black")
plt.plot(ts, theo_pdf, 'b-', linewidth=2, label="jMarkov PDF")

plt.xlabel("Service Time")
plt.ylabel("Density")
plt.title("Comparison of Service Time Distributions")
plt.legend()
plt.grid(True, linestyle="--", alpha=0.6)
plt.tight_layout()
plt.show()



import numpy as np
import math
from jmarkov.phase.ctph import ctph

class moments_ctph2():
    """
    Implements a PH/PH/1 queue and computes steady state metrics 
    
    The PH/PH/1 queue has phase type interarrival times, with parameters (alpha,T) of size ma,
    phase type service times, with parameters (beta,S) of size ms, 
    and 1 server.

    The class builds a quasi-birth-death chain that models this queue and uses 
    the steady state probability distribution of the chain to compute
    measures of performance such as mean number of entities in the system,
    in queue, in service, and the mean time in the system, in queue, and in service.
    """

    # first moment
    m1:float

    # second moment
    m2:float

    # phase-type adjusted to have moments m1 and m2
    PH:ctph

    
    # initializer 
    def __init__(self, m1:float, m2:float):
        """
        Creates a PH/PH/1 queue with 1 server, IAT inter-arrival times and ST service times
        """
        self.m1=m1
        self.m2=m2
       
        if m1 < 0 or m2 < 0:
            raise ValueError(f'The moments m1 ({m1}) and m2({m2}) must be positive')   
       

    def get_ph(self, verbose=False)->ctph:
        """
        Obtains the 2-phase PH representation that has the 2 moments requested
        """
        cx2 = self.m2/(self.m1*self.m1) - 1
        print(cx2)
        if cx2 < 1:
            print('cx<1')
            k = math.ceil(1/cx2)
            phi = k/self.m1
            alpha = np.zeros(k)
            alpha[0] = 1
            T = np.zeros((k,k))
            for i in range(k-1):
                T[i,i] = -phi
                T[i,i+1] = phi
            T[k-1,k-1] = -phi
            self.PH = ctph(alpha,T)
        else:
            print('cx>=1')
            # fix 3rd moment
            m12 = self.m1*self.m1
            m13 = m12*self.m1
            #m3 = 1.5*m13*(1+cx2)*(1+cx2)
            m3 = 2*m13*(1+cx2)*(1+cx2)
            #m3=6*m13*cx2
            #m3=6*m13*cx2*2
                
            d = 2*self.m1*self.m1 - self.m2
            c = 3*self.m2*self.m2-2*self.m1*m3
            b = 3*self.m1*self.m2-m3
            a= b*b-6*c*d

            print(a,b,c,d)
            
            lda1=lda2=p=0
            if abs(c)>1e-6:
                if c > 0:
                    print('c>0')
                    sqa=math.sqrt(a)
                    p =(-b+6*self.m1*d+sqa)/( b+sqa )
                    lda1=(b-sqa)/c
                    lda2=(b+sqa)/c
                else:
                    print('c<0')
                    sqa=math.sqrt(a)
                    p = (b-6*self.m1*d+sqa)/( -b+sqa )
                    lda1=(b+sqa)/c
                    lda2=(b-sqa)/c
            else:
                print("Adjusted to exponential")
                p=0
                lda2=1/self.m1
            print(p, lda1, lda2)
            alpha = np.array([p, 1-p])    
            T = np.array([
                [-lda1, 0],
                [0, -lda2]
            ])
            print(alpha)
            print(T)
            self.PH = ctph(alpha,T)
        return self.PH


#!/usr/bin/env python

import sys
import os.path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.phph1 import phph1
from jmarkov.phase.ctph import ctph

alpha = np.array([0.9, 0.1])
T = np.array([[-2, 1], [0, -3]])
IAT = ctph(alpha, T)
beta = np.array([0.3, 0.7])
S = np.array([[-5, 2], [0, -4]])
ST = ctph(beta, S)
q = phph1(IAT,ST) 
q._solve_mc(verbose=True)


print("Distributions of the number of entities in system:")
print(q.number_entities_dist())

L = q.mean_number_entities()
print("Mean number of entities in system:")
print(L)

Lq = q.mean_number_entities_queue()
print("Mean number of entities in queue:")
print(Lq)

W = q.mean_time_system()
print("Mean time in the system:")
print(W)

Wq = q.mean_time_queue()
print("Mean time in queue:")
print(Wq)

WT = q.wait_time_dist()
print("Waiting time distribution:")
print(f'alpha: {WT.alpha}')
print(f'T: {WT.T}')

rango = np.arange(0,1,0.001)*3*W

WT_pdf = pd.DataFrame(columns=['x','pdf'])
WT_cdf = pd.DataFrame(columns=['x','cdf'])
for x in rango:
    new_pdf = pd.DataFrame([[x, WT.pdf(x)]], columns=['x','pdf'])
    new_cdf = pd.DataFrame([[x, WT.cdf(x)]], columns=['x','cdf'])
    #print(new_pdf.head())
    WT_pdf = pd.concat([WT_pdf, new_pdf])
    WT_cdf = pd.concat([WT_cdf, new_cdf])


WT_pdf.plot(x='x', y='pdf')
plt.xlabel('x')
plt.ylabel('Waiting time pdf')
plt.show()

WT_cdf.plot(x='x', y='cdf')
plt.xlabel('x')
plt.ylabel('Waiting time cdf')
plt.show()


RT = q.resp_time_dist()
print("Response time distribution:")
print(f'alpha: {RT.alpha}')
print(f'T: {RT.T}')

mean_RT = RT.expected_value()
print(f'mean RT: {mean_RT}')
rango = np.arange(0,1,0.001)*3*mean_RT

RT_pdf = pd.DataFrame(columns=['x','pdf'])
RT_cdf = pd.DataFrame(columns=['x','cdf'])
for x in rango:
    new_pdf = pd.DataFrame([[x, RT.pdf(x)]], columns=['x','pdf'])
    new_cdf = pd.DataFrame([[x, RT.cdf(x)]], columns=['x','cdf'])
    #print(new_pdf.head())
    RT_pdf = pd.concat([RT_pdf, new_pdf])
    RT_cdf = pd.concat([RT_cdf, new_cdf])


RT_pdf.plot(x='x', y='pdf')
plt.xlabel('x')
plt.ylabel('Response time pdf')
plt.show()

RT_cdf.plot(x='x', y='cdf')
plt.xlabel('x')
plt.ylabel('Response time cdf')
plt.show()


import math
import numpy as np
from scipy import linalg
from scipy import sparse

class ctph():

    # number of phases
    n_phases:int=1

    # subgenerator matrix as 2d numpy array
    T:np.ndarray

    # initial phase vector as 1d numpy array
    alpha:np.ndarray
    
    # empty initializer 
    # chain with a single state
    def __init__(self):
        self.n_phases = 1
        self.T = np.array([-1])
        self.alpha = np.array([1])

    # initializer with only the number of states
    def __init__(self,n:int):
        self.n_phases = n
        self.generator = -1*np.eye(n, dtype=float)
        self.alpha = np.zeros(n, dtype=float)
        self.alpha[0] = 1

    # initializer with both alpha and T
    def __init__(self, alpha:np.array, T:np.array):
        if T.shape[0]!=T.shape[1]:
            raise ValueError("The dimensions of the T matrix are incorrect. It must be a square matrix.")
        self.n_phases=T.shape[0]
        self.T = T
        shape = alpha.shape
        self.alpha = alpha
        if shape[0]!=self.n_phases:
            raise ValueError("The dimensions of alpha vector are incorrect. Its length must coincide with the size of T")
        if not self._check_sub_generator_matrix(T): #Lets check if transition matrix is logical (i.e the rows sum 0)
            raise ValueError("the rows of transition matrix do not sum 0 or the diagonal has non negative values")
        
        

    def _check_sub_generator_matrix(self, M:np.ndarray):
        #Check if a given rate matrix has the condition that every row sum <= 0, with at least one < 0 
        check1 = np.max(np.sum(M, axis = 1)) <= 1e-10 and np.min(np.sum(M, axis = 1)) < -1e-10 
        # Check if a given transition matrix has all diagonal elements non positive
        if check1 and np.all(np.diag(M)<0):
            return True
        else:
            return False

    def pdf(self,t:float) -> float:
        res = 0.0
        if t == 0:
            res = 1 - self.alpha.sum()
        elif t > 0:                 
            P=linalg.expm(self.T*t)  #exponentiate the diferential generator following the method of Mohy et al (https://eprints.maths.manchester.ac.uk/1300/1/alhi09a.pdf)
            res = -self.alpha@P@(self.T).sum(1)
        return res
    
    def cdf(self,t:float,stop:float=0,step:float=0) -> np.ndarray:
        P=linalg.expm(self.T*t)  #exponentiate the diferential generator following the method of Mohy et al (https://eprints.maths.manchester.ac.uk/1300/1/alhi09a.pdf)
        if stop == 0 or step == 0:
            probs= 1 - np.float64(sum(self.alpha@P))
            return np.array(probs)
        elif stop > t and step > 0:
            n = math.ceil((stop-t)/step)
            probs = np.zeros(n)
            probs[0] = 1 - np.float64(sum(self.alpha@P))
            Pstep=linalg.expm(self.T*step)
            Ptemp = P
            for i in range(1,n):
                Ptemp = Ptemp*Pstep
                probs[i] = 1 - np.float64(sum(self.alpha@Ptemp)) 
            return probs
        else:
            return np.array(0.0)

  
    def expected_value(self):
        # Compute the expected value as alpha*inv(T)*one
        return np.sum(self.alpha@np.linalg.inv(-self.T))




import numpy as np
from scipy import linalg
from scipy import sparse
from .markov_chain import markov_chain

class ctmc(markov_chain):
    """
    Implements a finite continuous-time Markov chain (CTMC) 
    
    The chain is defined by its number of states, states, and a 
    generator matrix. The class provides methods to compute both
    stationary and transient metrics, as well as to check the 
    chain properties (ergodicity). 
    """

    # number of states in string array form
    n_states:int=1

    # states in string array form
    states:np.array = np.array([1])

    # generator matrix as 2d numpy array
    generator:np.array
    
    # empty initializer 
    # chain with a single state
    def __init__(self):
        self.n_states = 1
        self.generator = np.array([0])

    # initializer with only the number of states
    def __init__(self,n:int):
        self.n_states = n
        self.generator = np.zeros((n,n), dtype=float)

    # initializer with a generator matrix
    def __init__(self,generator:np.array, states:np.array=[1]):
        """
        Creates a continuous-time Markov chain from its generator matrix
        """
        if not self._check_generator_matrix(generator):#Lets check if transition matrix is logical (i.e the rows sum 0)
            raise ValueError("the rows of generator matrix do not sum 0 or the diagonal has non negative values")
        self.n_states=generator.shape[0]
        self.generator = generator
        self.states = states


    def _check_generator_matrix(self,M:np.ndarray, tol=10e-14):
        """
        Checks that a matrix is a Markov chain generator
         
        Checks that all row sums are equal to zero, diagonal entries are non-positive, and 
        non-diagonal entries are non-negative
        """
        # Check if the sum of the rows of the transition matrix equals (or is sufficiently close) to zero 
        check1 = np.max(np.sum(M, axis = 1)) <= tol and np.min(np.sum(M,axis=1)) >= -tol
        # Check if a given transition matrix has all diagonal elements non positive
        if check1 and np.all(np.diag(M)<=0):
            return True
        else:
            return False


    def steady_state(self) -> np.ndarray:
        """
        Computes the steady state distribution of the continuous-time Markov chain

        Computes the steady state probability distribution by replacing one of the 
        matrix equations with a normalizing equation that ensures the result is a 
        probability distribution. 

        Returns the stationary probability distribution in array form
        """

        # 
        if self.generator.any():
            # sets A to be Q and replace first column with ones for the normalizing equation
            A = self.generator.copy()
            A[:, 0] =  1
            # sets b as the right hand vector with a 1 for the normalizing equation
            b = np.zeros(self.n_states, dtype=float)
            b[0] = 1
            # solves transposed system as pi is a row vector
            res = linalg.solve(A,b,transposed=True)
            return res
        else:    
            print("Empty generator matrix")
        return 0
    

    def transient_probabilities(self,t:float,alpha:np.ndarray) -> np.ndarray:
        """
        Computes the transient distribution at time t with initial state alpha

        Computes alpha*exp(Q*t)*ones to obtain the probability distribution
        at time t given the initial probability distribution alpha
        """
        shape=alpha.shape#lets get the shape of alpha vector
        if shape[0]!=self.n_states:
            raise ValueError("The dimensions of alpha vector are incorrect. It must be a vector 1xn_states")
        if not np.isclose(sum(alpha),1,1e-10,1e-10):
            raise ValueError("The alpha vector does not sum to 1. It should be an stochastic vector.")
        P=linalg.expm(self.generator*t)#exponentiate the diferential generator following the method of Mohy et al (https://eprints.maths.manchester.ac.uk/1300/1/alhi09a.pdf)
        probs=alpha@P
        return probs


    def first_passage_time(self, target:str):
        """
        Computes the expected first passage time to a target state from a start state.

        This method calculates the expected number of steps required for the Markov chain to reach
        the specified target state from the start state by creating a sub-matrix of the generator matrix with the target removed.

        Returns the expected steps to reach the target state from any start state (except target) in array form        
        """
        #A copy of the transition matrix is created to solve the linear system
        m=self.generator.copy()
        
        # Matrix of ones with n-1 rows is created
        u=np.full([self.n_states-1,1],1)

        #The column and row corresponding to the target state are deleted from the transition matrix 
        m=np.delete(m,target, axis=0)
        m=np.delete(m,target, axis=1)

        t = linalg.solve(-m,u,transposed=False)
        return t

    def occupation_time(self,T,Epsilon=0.00001):
        """
        Computes the expected occupation time in each state until time T.

        This method computes the expected time the chain spends in each state,
        from time 0 until time T. To this end it uses the embedded matrix and 
        the uniformization method. 

        Returns the expected time in each state from 0 to T   
        """
        n=self.n_states
        m=self.generator.copy()
        def P_from_R(n,m):            
            # Inicializamos el valor máximo con el primer elemento de la diagonal
            Vector_ri=np.diagonal(m)   
            #Hallamos la r como el máximo de las r_i
            r=max(-Vector_ri)
            return r 
        def creacion_P(n,m):
            r= P_from_R(n,m)
            #Hallamos la matriz P^        
            h=np.zeros((n,n))
            for i in range(n):
                for j in range(n):
                    if i==j:
                        h[i][j]=1+(m[i][j]/r)
                    else:
                        h[i][j]=m[i][j]/r
            return h 
        r=P_from_R(n,m)
        P=creacion_P(n,m)
        i=np.identity(n)
        A=P
        k=0
        yek=np.exp(-r*T)
        ygk=1-yek
        suma=ygk
        B=ygk*i
        while suma/r < T-Epsilon:
            k=k+1
            yek=yek*(r*T)/k
            ygk=ygk-yek
            B=B+ygk*A
            A=A@P
            suma=suma+ygk
        return (B/r)
    

    def is_irreducible(self):
        """
        Checks if the given continous-time Markov Chain is irreducible.
        
        This method determines if the continous-time Markov Chain is irreducible by checking if, starting in
        any state, it is possible to reach any other state in a sequence of transitions.

        Returns a boolean
        """
        # the finite case:

        # We bring the generator matrix and replace all the diagonal elements for 0
        Q = np.copy(self.generator)
        np.fill_diagonal(Q,0)
        # We check if all the components in the chain are strongly connected
        if sparse.csgraph.connected_components(Q, directed=True,connection='strong',return_labels=False)==1:
            return True
        else:
            return False

    def is_ergodic(self):
        """
        Checks if a given continous-time Markov Chain is ergodic.
        
        Given that a finite continous-time Markov Chain is ergodic if it is irreducible, this
        method checks if it is irreducible to determine if the provided chain is ergodic or not.
        """
        # We check if the chain is irreducible
        if self.is_irreducible():
            return True
        else:
            return False
    
    def absorbtion_times(self,target:int,start=None):
        """
        Computes the mean time spent in state target before absorption, starting from state start. 

        If the chain is absorbing, both start and target must be transient. 
        If the chain is ergodic, this is the same as the mean first passage time 
        to state target, starting from state start. 
        """
        U = self.generator
        target = np.where(self.states == target)
        start = np.where(self.states == start) if start is not None else None
        if self.is_ergodic():
            U = np.delete(U, target, axis=0)
            U = np.delete(U, target, axis=1)
            # check that start is a transient state (must be different than target)
            if start != target:
                mat_Abs = np.linalg.inv(-U)
                
                return mat_Abs[start,]
            else:
                return "start should be different than target"
        else:
            # check that both start and end states are transient
            absorbing_states = np.where(np.all(U == 0, axis=1))[0]
            transient_states = np.setdiff1d(np.arange(U.shape[0]), absorbing_states)
            if np.isin(target, transient_states).all() and np.isin(start, transient_states).all():
                U = U[np.ix_(transient_states, transient_states)]
                mat_Abs = np.linalg.inv(-U)
                return mat_Abs[start,target]
            else:
                return "start and target should be transient"
            
    def absorbtion_probabilities(self, target:int, start=None):
        """
        Computes the probability of being absorbed by state target, starting from state start. 

        The chain must not be ergodic for the calculation to make sense.
        """
        if not self.is_ergodic():
            U = self.generator
            absorbing_states = np.where(np.all(U == 0, axis=1))[0]
            transient_states = np.setdiff1d(np.arange(U.shape[0]), absorbing_states)
            if np.isin(target, absorbing_states).all() and np.isin(start, transient_states).all():
                U_ = U[np.ix_(transient_states, transient_states)]
                V = U[np.ix_(transient_states, absorbing_states)]
                mat_Probs = np.matmul(np.linalg.inv(-U_),V)
                return mat_Probs[start,np.where(absorbing_states==target)]
        else:
            return "the chain shouldn't be ergodic"



import numpy as np
from scipy import linalg
from scipy import sparse
from .markov_chain import markov_chain
import math

class dtmc(markov_chain):
    """
    Implements a finite discrete-time Markov chain (CTMC) 
    
    The chain is defined by its number of states, states, and a 
    transition matrix. The class provides methods to compute both
    stationary and transient metrics, as well as to check the 
    chain properties (aperiodicity and ergodicity). 
    """

    # number of states in string array form
    n_states:int=1

    # states in string array form   
    states:np.array = np.array([1])

    # transition matrix as 2d numpy array
    transition_matrix:np.array
    
    # empty initializer 
    # chain with a single state
    def __init__(self):
        """
        Creates a continuous-time Markov chain from its transition matrix
        """
        self.n_states = 1
        self.transition_matrix = np.array([1])

    # initializer with only the number of states
    def __init__(self,n:int):
        self.n_states = n
        self.transition_matrix = np.eye(n, dtype=float)

    # initializer with a transition matrix
    def __init__(self,transition_matrix:np.array,states:np.array=[1]):
        if not self._check_transition_matrix(transition_matrix):#Lets check if transition matrix is logical (i.e the rows sum 1)
            raise ValueError("the rows of transition matrix do not sum 1 or has non positive values")
        self.n_states=transition_matrix.shape[0]
        self.transition_matrix = transition_matrix
        self.states = states

    def _check_transition_matrix(self, M:np.array):
        """
        Checks that a matrix is a Markov chain transition matrix
         
        Checks that all entries are non-negative and all row sums are equal to one
        """
        #Check if a given transition matrix has the condition that for every row the sum of all elements is equal to 1
        #Check if a given transition matrix has the condition that every element of the matrix is between 0 and 1
        if np.all(np.isclose(np.sum(M,axis=1),1,1e-5)) and np.all((M>=0) & (M<=1)):
            return True
        else:
            return False
        

    def steady_state(self) -> np.ndarray:
        """
        Computes the steady state distribution of the discrete-time Markov chain

        Computes the steady state probability distribution by replacing one of the 
        matrix equations with a normalizing equation that ensures the result is a 
        probability distribution. 

        Returns the stationary probability distribution in array form
        """
        # computes the steady state distribution
        if self.transition_matrix.any():
            # sets A to be I-P and replace first column with ones for the normalizing equation
            A = self.transition_matrix - np.eye(self.n_states, dtype=float)
            A[:, 0] =  1
            # sets b as the right hand vector with a 1 for the normalizing equation
            b = np.zeros(self.n_states, dtype=float)
            b[0] = 1
            # solves transposed system as pi is a row vector
            res = linalg.solve(A,b,transposed=True)
            return res
        else:    
            print("Empty transition matrix")
        return 0
    

    def transient_probabilities(self,n,alpha):
        """
        Computes the transient distribution at transition n with initial state alpha

        Computes alpha*(P^n)*ones to obtain the probability distribution
        at time/transition n given the initial probability distribution alpha
        """
         
        #First, lets verify n is integer
        if not isinstance(n,int):
            raise ValueError("The number of transitions n must be integer")
        #Now lets verify the dimensions of the vector alpha:
        try:
            shape=alpha.shape
        except:
            print("alpha vector must be a numpy array")
            return None
        if shape[0]!=self.n_states:
            raise ValueError("The dimensions of alpha vector are incorrect. It must be a vector 1xn_states")
        if not np.isclose(sum(alpha),1,1e-10,1e-10):
            raise ValueError("The alpha vector does not sum to 1. It should be an stochastic vector.")
        #Computes transient_matrix**n
        matrix_n=np.linalg.matrix_power(self.transition_matrix,n)
        vector=alpha@matrix_n
        return vector


    def first_passage_time(self, target:str):
        """
        Computes the expected first passage time to a target state from a start state.

        This method calculates the expected number of steps required for the Markov chain to reach
        the specified target state from the start state by creating a sub-matrix of the transition matrix with the target removed.

        Returns the expected steps to reach the target state from any start state (except target) in array form        
        """
        #The transition matrix and the number of states are brought
        n=self.n_states 
        M=self.transition_matrix.copy()
        # The identity matrix is created with size n-1 states
        i=np.identity(n-1)
        # Matrix of ones with n-1 rows is created
        u=np.full([n-1,1],1)
        # The column and row corresponding to the target state are eliminated from the transition matrix
        p=np.delete(M,target, axis=0)
        p=np.delete(p,target, axis=1)
        
        t=np.matmul(np.linalg.inv(i-p),u)
        return t 

    def occupation_time(self, n:int):
        """
        Computes the expected occupation time in n steps.

        This method calculates the expected time that the discrete-time Markov chain will remain in each state, from all starting states.


        Returns the occupation matrix in array form  
        """
        # Create an identity matrix with the same shape of transition matrix
        occupation=np.eye(self.transition_matrix.shape[0],self.transition_matrix.shape[1])
        # Iterate for each step
        for i in range(1,n+1):
            # Calculate the occupation time for each step, and update the occupation matrix
            occupation+=np.linalg.matrix_power(self.transition_matrix,i)
        return occupation
    
    def period(self):
        """
        Computes the period of a discrete-time Markov Chain.
        
        This method calculates the period of a discrete-time Markov Chain by iterating a total of N steps, where N represents
        the number of states that the chain has. In each k-th iteration, the chain is elevated to the k-th power to compute the
        transient probabilities in the k-th step. If one of the elementsin in the diagonal of the resulting matrix is greater than 0,
        the chain can potentially have a period k. The period of the matrix will be the Greatest Common Divisor (GCD) between the
        potential new period of the matrix in the k-th iteration and the current GCD before the k-th iteration.

        Returns the period of the chain as an int
        """
        
        #Initializes the transition matrix and the greatest common divisor for the periodicity calculations
        P_k = self.transition_matrix
        gcd = 0
        
        #Loop to elevate the transition matrix to the k th power and evaluate its periodicity:
        for k in range(2, (self.n_states)+1):
            P_k = np.dot(self.transition_matrix, P_k)
            #If the transition matrix to the k-th power has a non-zero element in its diagonal, the chain could have a period k:
            if np.any(np.diagonal(P_k) != 0):
                #GCD between the current GCD for the period of the matrix and the potential new GCD
                gcd = math.gcd(gcd, k)
            #If the GCD is 1, there is no need to keep looking for the period, the chain is aperiodic:
            if gcd == 1:
                break
        return gcd
    
    def is_irreducible(self):
        """
        Checks if the given discrete-time Markov Chain is irreducible.
        
        This method determines if the discrete-time Markov Chain is irreducible by checking if, starting in
        any state, it is possible to reach any other state in a sequence of transitions.

        Returns a boolean
        """
        # Brings a copy of the transition matrix
        P = np.copy(self.transition_matrix)
        # Checks if all the states in the Markov chain are strongly connected
        if sparse.csgraph.connected_components(P, directed=True,connection='strong',return_labels=False)==1:
            return True
        else:
            return False

    def is_ergodic(self):
        """
        Checks if a given discrete-time Markov Chain is ergodic.
        
        Given that a finite discrete-time Markov Chain is ergodic if it is aperiodic and irreducible, this
        method checks if both conditions are met to determine if the provided chain is ergodic or not.
        
        Returns a boolean
        """
        # Checks if the chain is irreducible and if it has a period of 1
        if self.period() == 1 & self.is_irreducible() == True:
            return True
        else:
            return False

    def absorbtion_times(self,target:int,start=None):
        """
        Computes the mean time spent in state target before absorption, 
        starting from state start. 

        If the chain is absorbing, both start and target must be transient. 
        If the chain is ergodic, this is the same as the mean first passage time 
        to state target, starting from state start. 
        """
        P = self.transition_matrix

        if self.is_ergodic():
            Q = np.delete(P, target, axis=0)
            Q = np.delete(Q, target, axis=1)
            # check that start is a transient state (must be different than target)
            if start != target:
                I = np.eye(Q.shape[0])
                mat_Abs = np.linalg.inv(I-Q)
                return mat_Abs[start,]
            else:
                return "start should be different than target"
        else:
            # check that both start and end states are transient
            absorbing_states = np.where( (np.sum(P == 1, axis=1) == 1) & (np.all((P == 0) | (P == 1), axis=1)) )[0]
            transient_states = np.setdiff1d(np.arange(P.shape[0]), absorbing_states)
            if np.isin(target, transient_states).all() and np.isin(start, transient_states).all():
                Q = P[np.ix_(transient_states, transient_states)]
                I = np.eye(len(transient_states))
                mat_Abs = np.linalg.inv(I-Q)
                return mat_Abs[start,target]
            else:
                return "start and target should be transient"
            
    def absorbtion_probabilities(self, target:int, start=None):
        """
        Computes the probability of being absorbed by state target, 
        starting from state start. 

        The chain must not be ergodic for the calculation to make sense.
        """
        if not self.is_ergodic():
            P = self.transition_matrix
            absorbing_states = np.where( (np.sum(P == 1, axis=1) == 1) & (np.all((P == 0) | (P == 1), axis=1)) )[0]
            transient_states = np.setdiff1d(np.arange(P.shape[0]), absorbing_states)
            if np.isin(target, absorbing_states).all() and np.isin(start, transient_states).all():
                Q = P[np.ix_(transient_states, transient_states)]
                I = np.eye(len(transient_states))
                R = P[np.ix_(transient_states, absorbing_states)]
                mat_Probs = np.matmul(np.linalg.inv(I-Q),R)
                return mat_Probs[start,np.where(absorbing_states==target)]
        else:
            return "the chain shouldn't be ergodic"







#!/usr/bin/env python
import sys
import os.path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.phase.fit.moments_ctph2 import moments_ctph2
from jmarkov.phase.ctph import ctph

m1 = 1
cx2 = 2
m2 = m1*m1*(cx2+1)

fitter = moments_ctph2(m1,m2)
PH = fitter.get_ph()
print(PH.alpha)
print(PH.T)


#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.queue.phph1 import phph1
from jmarkov.phase.ctph import ctph

# arrival rate
INTER_ARRIVAL_TIME = 5
lda = 1/INTER_ARRIVAL_TIME
alpha = np.array([1])
T = np.array([[-lda]])
IAT = ctph(alpha, T)

# paramters
lambda1 = 1.3862
mu1 = 3.8158
N = 20

# job service time distribution
mean_task_service_time = (1/lambda1)+(1/mu1)
num_tasks = N
task_service_rate = 1/mean_task_service_time
beta = np.zeros(num_tasks+1)
beta[0] = 1
S = np.zeros((num_tasks+1,num_tasks+1))
for i in range(num_tasks):
    if i==0:
        S[i,i] = -lambda1
        S[i,i+1] = lambda1
    else:
      #S[i,i] = -(num_tasks-i+1)*task_service_rate
      #S[i,i+1] = (num_tasks-i+1)*task_service_rate
      S[i,i] = -(num_tasks-i+1)*mu1
      S[i,i+1] = (num_tasks-i+1)*mu1



S[num_tasks,num_tasks] = -mu1
print(f'beta: {beta}')
print(f'S: {S}')
ST = ctph(beta, S)
q = phph1(IAT,ST)

print("Distributions of the number of entities in system:")
print(q.number_entities_dist())

L = q.mean_number_entities()
print("Mean number of entities in system:")
print(L)

Lq = q.mean_number_entities_queue()
print("Mean number of entities in queue:")
print(Lq)

W = q.mean_time_system()
print("Mean time in system:")
print(W)

Wq = q.mean_time_queue()
print("Mean time in queue:")
print(Wq)

WT = q.wait_time_dist()
print("Waiting time distribution:")
print(f'alpha: {WT.alpha}')
print(f'T: {WT.T}')




import numpy as np

class mg1():
    """
    Implements an M/G/1 queue and computes several performance metrics 
    
    The M/G/1 queue has exponential interarrival times, with rate arr_rate,
    and one server with service times that follow any probability distribution, 
    specified by its mean ser_mean and variances ser_var.

    The class employs the Pollaczek–Khinchine formula to compute measures
    of performance such as mean number of entities in the system, in queue, in service, 
    and the mean time in the system, in queue, and in service, as well, as server utilization.
    """
    # arrival rate
    arr_rate:np.float64

    # service time mean
    ser_mean:np.float64

    # service time variance
    ser_var:np.float64

    # initializer 
    def __init__(self, arr_rate:np.float64, ser_mean:np.float64, ser_var:np.float64):
        """
        Creates an M/G/1 queue with 1 server, arr_rate arrival rate, 
        ser_mean service time mean rate and ser_var service time variance
        """
        self.arr_rate = arr_rate
        self.ser_mean = ser_mean
        self.ser_var = ser_var

    def utilization(self)->np.float64:
        """
        Computes the mean server utilization
        
        The mean server utilization is computed as the ratio between the arrival rate
        and the service rate times the number of servers
        """
        rho = self.arr_rate*self.ser_mean
        return rho

    def mean_number_entities_queue(self)-> np.float64:
        """
        Computes the mean number of entities in queue in steady state
        """
        
        if self.is_stable:
            rho = self.utilization()
            return (((self.arr_rate**2)*(self.ser_var**2)) + (rho**2))/(2*(1-rho))
        else:
            print('Unstable queue')
            return 0
        
    def mean_number_entities_service(self)-> np.float64:
        """
        Computes the mean number of entities in service in steady state
        """
        if self.is_stable:
            return self.arr_rate*self.ser_mean
        else:
            print('Unstable queue')
            return 0
    
    def mean_number_entities(self)-> np.float64:
        """
        Computes the mean number of entities in the system in steady state
        """
        if self.is_stable:
            return self.mean_number_entities_queue() + self.mean_number_entities_service()
        else:
            print('Unstable queue')
            return 0
        
    def mean_time_system(self)-> np.float64:
        """
        Computes the mean time in the system in steady state 
        """
        if self.is_stable:
            return self.mean_number_entities()/self.arr_rate
        else:
            print('Unstable queue')
            return 0
        
    def mean_time_queue(self)->np.float64:
        """
        Computes the mean time in the queue in steady state
        """
        if self.is_stable:
            return self.mean_number_entities_queue()/self.arr_rate
        else:
            print('Unstable queue')
            return 0
        
    def mean_time_service(self)->np.float64:
        """
        Computes the mean time in service
        """
        if self.is_stable:
            return self.mean_number_entities_service()/self.arr_rate
        else:
            print('Unstable queue')
            return 0

    def is_stable(self)-> bool:
        """
        Returns True if the queue is stable, False otherwise
        
        This queue is stable if the arrival rate is smaller than the 
        maximum service rate
        """
        return self.arr_rate < 1/self.ser_mean

    

